﻿using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using System.IO;
using System.Collections.Generic;

namespace GGJ_Game
{
    class Level : IDisposable
    {
        public const string LegName = "item/leg";
        public const string ArmName = "item/arm";
        public const string EyeName = "item/eye";
        public const string BrainName = "item/brain";
        public const string BackgroundBox = "stuff/kiste00";
        public const string BackgroundTable = "stuff/tisch00";
        public const string Exit = "stuff/exit";

        private Texture2D backgroundLayers;

        private Player player;

        public ContentManager ContentManager
        {
            get { return contentManager; }
            set { contentManager = value; }
        }
        ContentManager contentManager;
        GraphicsDevice graphicsDevice;

        private Tile[,] tiles;
        public Vector2 tileSize;

        public List<Collectable> Collectables
        {
            get { return collectables; }
        }
        List<Collectable> collectables = new List<Collectable>();
        public List<Collectable> BoxObjects
        {
            get { return boxObjects; }
        }
        List<Collectable> boxObjects = new List<Collectable>();
        List<Collectable> hiddenBoxObjects = new List<Collectable>();   // some objects are hidden for the player and appear if he gets his eye upgrade
        List<Tile> hiddenTiles = new List<Tile>();
        List<Collectable> backgroundObjects = new List<Collectable>();
        public Collectable ExitObject
        {
            get { return exitObject; }
        }
        Collectable exitObject;

        public int LevelHeight
        {
            get { return tiles.GetLength(1); }
        }

        public int LevelWidth
        {
            get { return tiles.GetLength(0); }
        }

        public bool HasReachedExit
        {
            get { return hasReachedExit; }
        }
        bool hasReachedExit = false;

        // misc
        private int randomObjects = 6;
        private Random random = new Random();
        private List<string> backgroundObjectNames = new List<string>();
        private List<string> boxesObjectNames = new List<string>();

        public Game_GGJ Game
        {
            get { return game; }
            set { game = value; }
        }
        Game_GGJ game;

        // Constructor
        public Level(IServiceProvider serviceProvider, Stream fileStream, int levelIndex, GraphicsDeviceManager graphicsDeviceManager)
        {
            // New content manager to load content
            contentManager = new ContentManager(serviceProvider, "Content");
            graphicsDevice = graphicsDeviceManager.GraphicsDevice;

            backgroundLayers = contentManager.Load<Texture2D>("screens/background_0");

            // We need the box names inside the loadLevelFile()
            fillBackgroundObjectNames();

            loadLevelFile(fileStream);

            // background objects
            // Currently we don't use random objects
            //placeRandomBackgroundObjects();
        }

        // Load the level file from harddisk
        private void loadLevelFile(Stream fileStream)
        {
            int linewidth;
            List<string> lines = new List<string>();
            using (StreamReader reader = new StreamReader(fileStream))
            {
                string line = reader.ReadLine();
                linewidth = line.Length;
                while (line != null)
                {
                    lines.Add(line);
                    if (line.Length != linewidth)
                    {
                        throw new Exception(String.Format("Zeilen abstand in zeile {0} unterscheidet sich von den anderen.", lines.Count));
                    }
                    line = reader.ReadLine();
                }
            }

            // Allocate the tiles
            tiles = new Tile[linewidth, lines.Count];

            // Recalculate the sprite size to support fixed screen size            
            tileSize = new Vector2(graphicsDevice.ScissorRectangle.Width / LevelWidth, graphicsDevice.ScissorRectangle.Height / LevelHeight);

            for (int y = 0; y < LevelHeight; ++y)
            {
                for (int x = 0; x < LevelWidth; ++x)
                {
                    char charType = lines[y][x];
                    tiles[x, y] = LoadTile(x, y, charType);
                }
            }
        }

        public Tile LoadTile(int x, int y, char charType)
        {
            switch (charType)
            {
                // Top-Right block
                case '9': return new Tile(x, y, ContentManager.Load<Texture2D>("tiles/BlockA6"), TileBlocking.Blocking_Full, tileSize);
                // Hidden top-right block
                case ')': hiddenTiles.Add(new Tile(x, y, ContentManager.Load<Texture2D>("tiles/BlockA6"), TileBlocking.Blocking_Full, tileSize));
                    break;
                // Top block
                case '8': return new Tile(x, y, ContentManager.Load<Texture2D>("tiles/BlockA1"), TileBlocking.Blocking_Full, tileSize);
                // Hidden top block
                case '(': hiddenTiles.Add(new Tile(x, y, ContentManager.Load<Texture2D>("tiles/BlockA1"), TileBlocking.Blocking_Full, tileSize));
                    break;
                // Top-left block
                case '7': return new Tile(x, y, ContentManager.Load<Texture2D>("tiles/BlockA6"), TileBlocking.Blocking_Full, tileSize, true);
                // Hidden top-left block
                case '/': hiddenTiles.Add(new Tile(x, y, ContentManager.Load<Texture2D>("tiles/BlockA6"), TileBlocking.Blocking_Full, tileSize, true));
                    break;
                // Right block
                case '6': return new Tile(x, y, ContentManager.Load<Texture2D>("tiles/BlockA4"), TileBlocking.Blocking_Full, tileSize);
                // Hidden right block
                case '&': hiddenTiles.Add(new Tile(x, y, ContentManager.Load<Texture2D>("tiles/BlockA4"), TileBlocking.Blocking_Full, tileSize));
                    break;
                // Normal block
                case '5': return new Tile(x, y, ContentManager.Load<Texture2D>("tiles/BlockA0"), TileBlocking.Blocking_Full, tileSize);
                // Hidden normal block
                case '%': hiddenTiles.Add(new Tile(x, y, ContentManager.Load<Texture2D>("tiles/BlockA0"), TileBlocking.Blocking_Full, tileSize));
                    break;
                // Left block
                case '4': return new Tile(x, y, ContentManager.Load<Texture2D>("tiles/BlockA4"), TileBlocking.Blocking_Full, tileSize, true);
                // Hidden left block
                case '$': hiddenTiles.Add(new Tile(x, y, ContentManager.Load<Texture2D>("tiles/BlockA4"), TileBlocking.Blocking_Full, tileSize, true));
                    break;
                // Left-bottom block
                case '3': return new Tile(x, y, ContentManager.Load<Texture2D>("tiles/BlockA5"), TileBlocking.Blocking_Full, tileSize);
                // Hidden left-bottom block
                case 'Q': hiddenTiles.Add(new Tile(x, y, ContentManager.Load<Texture2D>("tiles/BlockA5"), TileBlocking.Blocking_Full, tileSize));
                    break;
                // Bottom block
                case '2': return new Tile(x, y, ContentManager.Load<Texture2D>("tiles/BlockA2"), TileBlocking.Blocking_Full, tileSize);
                // Hidden bottom block
                case '\"': hiddenTiles.Add(new Tile(x, y, ContentManager.Load<Texture2D>("tiles/BlockA2"), TileBlocking.Blocking_Full, tileSize));
                    break;
                // Left-bottom block
                case '1': return new Tile(x, y, ContentManager.Load<Texture2D>("tiles/BlockA5"), TileBlocking.Blocking_Full, tileSize, true);
                // Hidden left-bottom block
                case '!': hiddenTiles.Add(new Tile(x, y, ContentManager.Load<Texture2D>("tiles/BlockA5"), TileBlocking.Blocking_Full, tileSize, true));
                    break;
                // Top-down hybrid block
                case '0': return new Tile(x, y, ContentManager.Load<Texture2D>("tiles/BlockA3"), TileBlocking.Blocking_Full, tileSize);
                // Hidden top-down hybrid block
                case '=': hiddenTiles.Add(new Tile(x, y, ContentManager.Load<Texture2D>("tiles/BlockA3"), TileBlocking.Blocking_Full, tileSize));
                    break;

                  // Item
                case 'A':
                    collectables.Add(new Collectable(x, y, ContentManager, LegName, tileSize));
                    break;
                case 'B':
                    collectables.Add(new Collectable(x, y, ContentManager, ArmName, tileSize));
                    break;
                case 'C':
                    collectables.Add(new Collectable(x, y, ContentManager, EyeName, tileSize));
                    break;
                case 'D':
                    collectables.Add(new Collectable(x, y, ContentManager, BrainName, tileSize));
                    break;
                case 'E':
                    exitObject = new Collectable(x, y-1, ContentManager, Exit, tileSize, 1, true);
                    break;
                    // Hidden plattform
                case '_': hiddenTiles.Add(new Tile(x, y, ContentManager.Load<Texture2D>("tiles/PlattformA"), TileBlocking.Blocking_Half, tileSize));
                    break;
                case '-': return new Tile(x, y, ContentManager.Load<Texture2D>("tiles/PlattformA"), TileBlocking.Blocking_Half, tileSize);
                    // Hidden disapering block
                case ':': hiddenTiles.Add(new Tile(x, y, null, TileBlocking.Blocking_No, tileSize)); 
                    // return normal block
                    goto case '5';
                    // Box
                case 'k':
                    boxObjects.Add(new Collectable(x, y, contentManager, getRandomBoxObjectName(), tileSize, 2));
                    break;
                    // Hidden box
                case 'K':
                    hiddenBoxObjects.Add(new Collectable(x, y, contentManager, getRandomBoxObjectName(), tileSize, 2));
                    break;
                // Create player
                case 'p':
                    {
                        player = new Player(this, new Vector2(x, y) * tileSize);
                        break;
                    }
            }

            // No Gfx
            return new Tile(x, y, null, TileBlocking.Blocking_No, tileSize);
        }

        private void fillBackgroundObjectNames()
        {
            this.boxesObjectNames.Add("stuff/kiste00");
            this.boxesObjectNames.Add("stuff/kiste00a");
            this.boxesObjectNames.Add("stuff/kiste00b");
            this.boxesObjectNames.Add("stuff/kiste00c");
            this.boxesObjectNames.Add("stuff/kiste00d");
            this.boxesObjectNames.Add("stuff/kiste00e");
            this.boxesObjectNames.Add("stuff/kiste00f");

            this.backgroundObjectNames.Add("stuff/tisch00");
            this.backgroundObjectNames.Add("stuff/tisch00a");
            this.backgroundObjectNames.Add("stuff/tisch00b");
            this.backgroundObjectNames.Add("stuff/tisch00c");
            this.backgroundObjectNames.Add("stuff/tisch00d");
        }
        
        private void placeRandomBackgroundObjects()
        {
            for (int i = 0; i < randomObjects; ++i)
            {
                for (int placementTry = 0; placementTry < 20; ++placementTry)
                {
                    int xPos = random.Next(LevelWidth-2);
                    int yPos = random.Next(LevelHeight - 1);

                    if (getCollision(xPos, yPos + 2) == TileBlocking.Blocking_Full)
                    {
                        // Found blocking tile, also check the right neighbor
                        if (getCollision(xPos + 1, yPos + 2) == TileBlocking.Blocking_Full)
                        {
                            // Good place to spawn random object
                            // TODO(mk) Add check for colliding with other background objects
                            string objectName = getRandomBackgroundObjectName();
                            this.backgroundObjects.Add(new Collectable(xPos, yPos, contentManager, objectName, tileSize, 2));
                            break;
                        }
                    }
                }
            }
        }

        private string getRandomBackgroundObjectName()
        {
            return backgroundObjectNames[random.Next(backgroundObjectNames.Count - 1)];
        }

        private string getRandomBoxObjectName()
        {
            return boxesObjectNames[random.Next(boxesObjectNames.Count - 1)];
        }

        double lastBoxUpdate = 0;
        public void Update(GameTime gameTime, KeyboardState keyboardState)
        {
            const double BoxUpdateInterval = 0.5;
            lastBoxUpdate += gameTime.ElapsedGameTime.TotalSeconds;

            player.Update(gameTime, keyboardState);

            foreach (Collectable colletc in collectables)
            {
                colletc.Update(gameTime);
            }

            // Boxes
            if (lastBoxUpdate >= BoxUpdateInterval)
            {
                lastBoxUpdate = 0;
                foreach (Collectable boxObject in boxObjects)
                {
                    boxObject.Update(gameTime);

                    // Reset flag to search for ground collision.
                    bool boxIsOnGround = false;

                    // Check for collision
                    TileBlocking tileBlocking = getCollision(boxObject.xPos, boxObject.yPos + 2);
                    if (tileBlocking == TileBlocking.Blocking_No)
                    {
                        tileBlocking = getCollision(boxObject.xPos + 1, boxObject.yPos + 2);
                    }
                    if (tileBlocking != TileBlocking.Blocking_No)
                    {
                        boxIsOnGround = true;
                    }

                    if (!boxIsOnGround)
                    {
                        boxObject.yPos += 1;
                    }
                }
            }

            // Play animation
            exitObject.Update(gameTime);

            if (player.HasEyes)
            {
                showHiddenBoxes();
                showHiddenTiles();
            }
            if (player.HasReachedExit)
            {
                // Tell the level to abort 
                this.hasReachedExit = true;
            }
        }

        public void UpdateHud(IngameHud ingameHud)
        {
            ingameHud.ShowLeg = player.HasLegs;
            ingameHud.ShowArm = player.HasArms;
            ingameHud.ShowEye = player.HasEyes;
            ingameHud.ShowBrain = player.HasBrain;
        }

        void showHiddenBoxes()
        {
            foreach (Collectable boxObject in hiddenBoxObjects)
            {
                boxObjects.Add(boxObject);
            }
            hiddenBoxObjects.Clear();
        }

        void showHiddenTiles()
        {
            foreach (Tile tile in hiddenTiles)
            {
                tiles[tile.xPos, tile.yPos] = tile;
            }
            hiddenTiles.Clear();
        }

        public void Dispose()
        {
            contentManager.Unload();
        }

        public void Draw(GameTime gameTime, SpriteBatch spriteBatch)
        {
            // Background
            spriteBatch.Draw(backgroundLayers, Vector2.Zero, Color.White);

            // Tiles
            DrawTiles(gameTime, spriteBatch);

            // Boxes
            foreach (Collectable collectc in boxObjects)
            {
                collectc.Draw(gameTime, spriteBatch);
            }

            // Items
            foreach (Collectable colletc in collectables)
            {
                colletc.Draw(gameTime, spriteBatch);
            }

            // Background items
            foreach (Collectable collect in backgroundObjects)
            {
                collect.Draw(gameTime, spriteBatch);
            }

            if (player.HasBrain)
            {
                exitObject.Draw(gameTime, spriteBatch, this.tileSize);
            }

            // Player
            player.Draw(gameTime, spriteBatch);
        }

        public void DrawTiles(GameTime gameTime, SpriteBatch spriteBatch)
        {
            for (int y = 0; y < LevelHeight; y++)
            {
                for (int x = 0; x < LevelWidth; x++)
                {
                    // Check for existing texture
                    Texture2D texture = tiles[x,y].Texture;
                    if (texture != null)
                    {
                        Vector2 drawPosition = new Vector2(x, y) * tileSize;
                        Rectangle blitRect = new Rectangle((int)drawPosition.X, (int)drawPosition.Y, (int)tileSize.X, (int)tileSize.Y);
                        SpriteEffects spriteeffects = (tiles[x,y].flip ? SpriteEffects.FlipHorizontally : SpriteEffects.None);

                        spriteBatch.Draw(texture, blitRect, null, Color.White, 0, new Vector2(), spriteeffects, 0);
                    }
                }
            }
        }

        public TileBlocking getCollision(int xPos, int yPos)
        {
            // stop left or right of the level
            if (xPos < 0 || xPos >= LevelWidth)
                return TileBlocking.Blocking_Full;
            if (yPos < 0 || yPos >= LevelHeight)
                return TileBlocking.Blocking_No;

            TileBlocking blockingType = tiles[xPos, yPos].TileBlockingType;
            if (blockingType == TileBlocking.Blocking_No && getCollisionWithBox(xPos, yPos) != null)
            {
                blockingType = TileBlocking.Blocking_Box;
            }
            return blockingType;
        }

        public Collectable getCollisionWithCollectable(int xPos, int yPos)
        {
            foreach (Collectable collectable in collectables)
            {
                if (collectable.xPos == xPos && collectable.yPos == yPos)
                {
                    return collectable;
                }
            }
            if (exitObject.xPos == xPos && exitObject.yPos == yPos)
            {
                return exitObject;
            }
            return null;
        }

        public Collectable getCollisionWithBox(int xPos, int yPos)
        {
            foreach (Collectable collectable in boxObjects)
            {
                if ((collectable.xPos == xPos || collectable.xPos + 1 == xPos) && (collectable.yPos == yPos || collectable.yPos + 1 == yPos))
                {
                    return collectable;
                }
            }
            return null;
        }

        public Rectangle getBounds(int xPos, int yPos)
        {
            return new Rectangle(xPos * (int)tileSize.X, yPos * (int)tileSize.Y, (int)tileSize.X, (int)tileSize.Y);
        }
    }
}
